#
# find_negative_cycle.py
# An implementation of the multiplication-based Bellman-Ford Algorithm for negative-cycle finding
# Author: Mihir Patil
#

import numpy as np
from bellman_ford import multiplicative_bellman_ford
from utils import get_prec


def find_negative_cycle(graph, graph_labels):
    """
    Finds negative cycles in the specified graph

    :param: graph The graph on which to operate.
    :param: graph_labels A ordered list of labels that correspond to the indices in the input graph.

    :return: If a negative cycle is found, returns a tuple with the negative cycle as a list,
             followed by the gain of pursuing the negative cycle. If no negative cycle is found,
             returns (None, None).
    """

    distance, predecessor = multiplicative_bellman_ford(graph)

    for i, node_a_weights in enumerate(graph):
        for j, weight in enumerate(node_a_weights):

            if weight is None:
                continue

            precision = min(get_prec(distance[i]), get_prec(weight), get_prec(distance[j]))

            dt_j = distance[j] if distance[j] == float('inf') else round(distance[j], precision)
            if round(distance[i] * weight, precision) < round(dt_j, precision):
                gain = 1.0
                negative_cycle = []
                negative_cycle.append(graph_labels[i])

                prev = i
                curr = int(predecessor[i])
                gain *= graph[prev][curr]

                while curr != i:
                    negative_cycle.append(graph_labels[int(curr)])
                    prev = curr
                    curr = int(predecessor[curr])
                    gain *= graph[prev][curr]

                negative_cycle.append(graph_labels[i])
                return negative_cycle, gain
    return None, None



import unittest
import numpy as np
from negative_cycles import find_negative_cycle


class TestNegativeCycles(unittest.TestCase):

    def _check(self, actual, expected):
        np.testing.assert_allclose(actual, expected)


    def _check_eq(self, actual, expected):
        assert actual == expected, 'Two expected {0}, but got {1}'.format(expected, actual)


    def _check_none(self, actual):
        np.testing.assert_equal(actual, None)


    def _check_array(self, actual, expected):
        np.testing.assert_array_equal(actual, expected)


    def test_small_graph(self):

        currency_list = np.array(['A', 'B', 'C'])
        graph = np.eye(len(currency_list))

        # A to B and back
        graph[0][1] = 1 / 1.5
        graph[1][0] = 1.5

        # A to C and back
        graph[0][2] = 10.0
        graph[2][0] = 1 / 10.0

        # B to C
        graph[1][2] = 16.0
        graph[2][1] = 1 / 16.0

        cycle, gain = find_negative_cycle(graph, currency_list)
        self._check_array(cycle, ['A', 'B', 'C', 'A'])
        self._check(gain, 1.0666666666666667)


    def test_four_node_graph(self):

        currency_list = np.array(['A', 'B', 'C', 'D'])
        graph = np.eye(len(currency_list))

        # A to B and back
        graph[0][1] = 1 / 1.09
        graph[1][0] = 1.09

        # A to C and back
        graph[0][2] = 6.48
        graph[2][0] = 1 / 6.48

        # B to C
        graph[1][2] = 7.08
        graph[2][1] = 1 / 7.08

        # A to D and back
        graph[0][3] = 66.027
        graph[3][0] = 1 / 66.027

        # B to D and back
        graph[1][3] = 72.197
        graph[3][1] = 1 / 72.197

        # C to D and back
        graph[2][3] = 10.191
        graph[3][2] = 1 / 10.191

        cycle, gain = find_negative_cycle(graph, currency_list)
        self._check_none(cycle)
        self._check_none(gain)


    def test_live_example_graph(self):

        currency_list = np.array(['A', 'B', 'C', 'D'])
        graph = np.array([[1.00000000e+00, 1.00000000e+00, 1.09110747e-02, 4.21229992e-02],
                          [1.00000000e+00, 1.00000000e+00, 1.87899286e+01, 7.40192450e+01],
                          [9.16500000e+01, 5.32200000e-02, 1.00000000e+00, None],
                          [2.37400000e+01, 1.35100000e-02, None,           1.00000000e+00]])

        cycle, gain = find_negative_cycle(graph, currency_list)
        self._check_array(cycle, ['B', 'C', 'A', 'B'])
        self._check(gain, 1722.0969)




#
# bellman_ford.py
#

import numpy as np
from utils import get_prec, is_int


def multiplicative_bellman_ford(graph):
    return bellman_ford(graph, is_multiplicative=True)


def additive_bellman_ford(graph):
    return bellman_ford(graph, is_multiplicative=False)


def bellman_ford(graph, is_multiplicative=False):
    """
    An implementation of the multiplication-based Bellman-Ford algorithm.

    :param: graph - The graph on which to operate. Should be a square matrix, where edges that don't
                    exist have value None
    :param: graph_labels - An ordered list of labels that correspond to the indices in the input
                           graph.
    :param: is_multiplicative - If this is True, performs multiplication-based Bellman Ford, where
                                the distances between two nodes are based on the smallest PRODUCT of
                                the edge weights between them. If it is False, then performs
                                addition-based Bellman Ford, where the distances between two nodes
                                are based on the SUM of the edge weights between them.

    :return: a tuple, where the zero-th item is the distance array output from the Bellman-Ford
             Algorithm, as well as the predecessor array to find paths
    """
    operator = (lambda x, y: x * y) if is_multiplicative else (lambda x, y: x + y)
    lowest_precision = float('inf')

    # Create a distance array with value infinity
    distance = np.zeros(len(graph)).astype(np.float64)
    distance.fill(float('inf'))
    distance[0] = 1.0 if is_multiplicative else 0

    # Create a predecessor array with value None
    predecessor = np.zeros(len(graph))
    predecessor.fill(-1)

    for _ in range(len(graph) - 1):

        # Iterate through all the vertices
        for i, node_a_weights in enumerate(graph):
            for j, weight in enumerate(node_a_weights):

                if weight is None:
                    continue

                # Handle precision development
                precision = min(get_prec(distance[i]), get_prec(weight), get_prec(distance[j]))
                lowest_precision = min(lowest_precision, precision)

                # Find the correct precision, compensating for infinity
                dt_j = distance[j] if distance[j] == float('inf') or is_int(distance[j]) \
                    else round(distance[j], precision)

                raw_new_dist = operator(distance[i], weight)
                new_dist = raw_new_dist if is_int(raw_new_dist) else round(raw_new_dist, precision)

                if new_dist < dt_j:
                    distance[j] = new_dist
                    predecessor[j] = i

    return distance, predecessor


import unittest
import numpy as np
from negative_cycles import additive_bellman_ford


class TestAdditiveBellmanFord(unittest.TestCase):

    def _check(self, actual, expected):
        np.testing.assert_allclose(actual, expected)


    def _check_eq(self, actual, expected):
        assert actual == expected, 'Two expected {0}, but got {1}'.format(expected, actual)


    def _check_none(self, actual):
        np.testing.assert_equal(actual, None)


    def _check_array(self, actual, expected):
        np.testing.assert_array_equal(actual, expected)


    def test_small_graph(self):

        graph = np.eye(3)

        # A to B and back
        graph[0][1] = 1
        graph[1][0] = 2

        # A to C and back
        graph[0][2] = 7
        graph[2][0] = 1

        # B to C and back
        graph[1][2] = 3
        graph[2][1] = 4

        distance, predecessor = additive_bellman_ford(graph)
        expected_distance = np.array([0, 1, 4])
        expected_predecessor = np.array([-1, 0, 1])

        self._check_array(distance, expected_distance)
        self._check_array(predecessor, expected_predecessor)


    def test_four_node_graph_with_negative_weight(self):

        graph = np.eye(4)

        # A to B and back
        graph[0][1] = 5
        graph[1][0] = 7

        # A to C and back
        graph[0][2] = 10
        graph[2][0] = 2

        # B to C and back
        graph[1][2] = 4
        graph[2][1] = 1

        # A to D and back
        graph[0][3] = 5
        graph[3][0] = 3

        # B to D and back
        graph[1][3] = 1
        graph[3][1] = -1

        # C to D and back
        graph[2][3] = 1
        graph[3][2] = 7

        distance, predecessor = additive_bellman_ford(graph)
        expected_distance = np.array([0, 4, 8, 5])
        expected_predecessor = np.array([-1, 3, 1, 0])

        self._check_array(distance, expected_distance)
        self._check_array(predecessor, expected_predecessor)




"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='negative-cycles',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.1.0',

    description='For Negative Cycle Detection',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/mnpatil17/negative_cycles/',

    # Author details
    author='Mihir Patil',
    author_email='mnpatil17@gmail.com',

    # Choose your license
    license='BSD 3-clause "New" or "Revised License"',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        'Programming Language :: Python :: 2.7',
    ],

    # What does your project relate to?
    keywords='negative-cycle bellman-ford',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=['numpy', 'nose'],

    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },
)